/*****************************************************************************
 * main.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * Copyright (C) 2000-2002 Chris Pinkham
 * Released under the terms of the GPL.
 * *NO WARRANTY*
 *
 * cpinkham@infi.net, cpinkham@bc2va.org
 * http://www4.infi.net/~cpinkham/gyach/
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <gtk/gtk.h>

#include "gyach.h"

#ifndef OS_WINDOWS
#	include <X11/Xlib.h>
#	include <gdk/gdkx.h>
#endif

#include <errno.h>

#include "interface.h"
#include "support.h"

#include "commands.h"
#include "callbacks.h"
#include "images.h"
#include "yahoochat.h"
#include "gyach_int.h"
#include "main.h"

/* 
 * Portions copyright David J. Binette from the yahmini source code
 * ALL RIGHTS RESERVED 
 */ 

/* ------------------------------------------------------------------------ */ 
 
SESSION session ,*sess = &session; /* for consistency */ 
 

typedef enum
{ Medium, Bold }
FontWeight;
typedef enum
{ Italic, Roman }
FontSlant;

static GdkFont *get_font(const char *family, const short size,
	const FontWeight weight, const FontSlant slant)
{
	GdkFont *font = NULL;
	char XLFD[1000];
	char my_font_name[129];
	char *ptr;
	char *from;
	char *to;

#ifndef OS_WINDOWS
	XFontStruct *fontOK = (XFontStruct *) NULL;
#endif

	DBG( 21, "get_font( '%s', %d, %d, %d )\n", family, size, weight, slant );

	if ( font ) {
		gdk_font_unref( font );
		font = NULL;
	}

	strcpy( my_font_name, font_name );
	strcpy(XLFD, "-*");			/* foundry */

	if (family && *family)
	{
		char *p = NULL, *lowerFace = NULL;

		strcat(XLFD, "-");
		p = lowerFace = strdup(family);
		while (p && *p)
		{
			int c = tolower(*p);

			*p = c;
			++p;
		}
		strcat(XLFD, lowerFace);
		free(lowerFace);
	}
	else
	{
		strcat(XLFD, "-times");	/* default family, I should really get this from the current theme */
	}

	switch (weight)
	{
		case Bold:
			strcat(XLFD, "-bold");	/* weight */
			if (( ptr = strstr( my_font_name, "medium" ))) {
				strcpy( ptr, "bold" );
				to = ptr + 4;
				from = ptr + 6;
				while( *from ) {
					*(to++) = *(from++);
				}
				*to = '\0';
			}
			break;

		case Medium:
		default:
			strcat(XLFD, "-medium");	/* weight */
			if (( ptr = strstr( my_font_name, "bold" ))) {
				from = strstr( font_name, "bold" ) + 4;
				strcpy( ptr, "medium" );
				to = ptr + 6;
				while( *from ) {
					*(to++) = *(from++);
				}
				*to = '\0';
			}
			break;
	}

	switch (slant)
	{
		case Italic:
			strcat(XLFD, "-i");	/* slant : italic */
			break;

		case Roman:
		default:
			strcat(XLFD, "-r");	/* slant : roman */
			break;
	}

	strcat(XLFD, "-normal");	/* set width */
	strcat(XLFD, "-*");			/* additional style (nil) */

	/* dunno why, but he appears to have had these next 2 fields reversed, it */
	/* works better this way. */
	/*
	if (size > -1)
	{
		char sizeS[10];

		sprintf(sizeS, "-%d", 10 * size);
		strcat(XLFD, sizeS);
	}
	else
	{
		strcat(XLFD, "-*");
	}
	*/

	{
		char sizeS[10];

		sprintf(sizeS, "-%d", font_size);
		strcat(XLFD, sizeS);
	}
	strcat(XLFD, "-*");			/* pixel size (don't care) */

	strcat(XLFD, "-*");			/* resolution X */
	strcat(XLFD, "-*");			/* resolution Y */
	strcat(XLFD, "-*");			/* proportional spacing */
	strcat(XLFD, "-*");			/* Average width */
	strcat(XLFD, "-*");			/* Charset registry */
	strcat(XLFD, "-*");			/* Charset encoding */

	/*
	printf( "XLFD: %s\n", XLFD );
	printf( "  my: %s\n", my_font_name );
	fflush( stdout );
	*/

	if ( ! show_fonts ) {
		strcpy( XLFD, my_font_name );
	}

	/* Test load the font.  If it fails, fontOK is null, and get on with
	 * life after giving the user an error.
	 * If it is successful, immediately unload the font, since we're using
	 * GTK+ here, mmmk?
	 */
#ifndef OS_WINDOWS
	fontOK = XLoadQueryFont(GDK_DISPLAY(), XLFD);

	if (fontOK == (XFontStruct *) NULL || (!fontOK))
	{
		/* Font was no good - you can't load this font. */
		font = NULL;
	}
	else
	{
		/* Remember to free this since XLoadQueryFont returns allocated memory */
		XFreeFont(GDK_DISPLAY(), fontOK);
		font = gdk_font_load(XLFD);
	}
#endif

	if ( ! font ) {
		strcpy( XLFD, "-*-times-medium-r-normal-*-14-*-*-*-*-*-*-*" );
		font = gdk_font_load( XLFD );
	}

	return font;
}

static GdkFont *getFont(int bold, int italic, int underline,
	const char *family, const char *sizeS)
{
	static GdkFont *font = NULL;
	FontWeight weight;
	FontSlant slant;
	short size = 0;

	DBG( 21, "getFont( %d, %d, %d, '%s', '%s' )\n",
		bold, italic, underline, family, sizeS );

	if (sizeS && *sizeS)
	{
		size = atoi(sizeS);
	}
	weight = bold ? Bold : Medium;
	slant = italic ? Italic : Roman;

	if (NULL != font)
	{
		gdk_font_unref(font);
		font = NULL;
	}

	if (NULL == (font = get_font(family, size, weight, slant)))
	{
		/* If the getting the actual font fails,
		   first try a few sizes smaller and
		   then try a blank family */
		short trySize = size - 1, endTrySize = (size - 4);

		if (family && *family)
		{
			while ((trySize > 0) &&
				(NULL == (font = get_font(family, trySize, weight, slant))) &&
				(trySize > endTrySize))
			{
				--trySize;
			}
		}

		if (NULL == font)
		{
			trySize = size;
			while ((trySize > 0) &&
				(NULL == (font = get_font(NULL, trySize, weight, slant))) &&
				(trySize > endTrySize))
			{
				--trySize;
			}
		}

	}

	if (NULL == font)
	{
		gdk_font_ref(font);
	}

	return (font);
}

static int bold = FALSE, italic = FALSE, underline = FALSE, url = FALSE;

void append_to_textbox(GtkWidget *window, GtkWidget *textbox, char *text)
{
	int i = 0;
	GdkColor colorspace;
	GdkColor *color = &colorspace;
	GdkFont *font = NULL;
	char face[1000] = "", size[1000] = "";
	int tmp_color;
	char *tmp_ptr;
	int scroll = FALSE;
	unsigned int r, g, b;
	unsigned int or, og, ob;

	DBG( 21, "append_to_textbox( %p, %p, '%s' )\n",
		window, textbox, text );

	*face = *size = '\0';

	strcpy( face, font_family );

	font = getFont(bold, italic, underline, face, size);

	if ( force_lowercase )
		lower_str( text );

	strip_junk( text );
	convert_tags( text );

	if (( ! textbox ) &&
		( ct_scroll_at_bottom())) {
		scroll = TRUE;
	}

	/* Reset all attributes - currently just color */
	r = g = b = 0;
	or = og = ob = 0;

	/* freeze before printing to speed up display, especially on blended */
	/* colors or for people who switch colors every letter */
	if ( ! textbox )
		ct_freeze();

	i = 0;
	while (text && i < strlen(text))
	{
		if (!strncmp(&text[i], YAHOO_FONT_TAG_START,
				strlen(YAHOO_FONT_TAG_START)))
		{
			int closeFound = FALSE;

			i += strlen(YAHOO_FONT_TAG_START);
			while (!closeFound)
			{
				if (!strncmp(&text[i], YAHOO_FONT_FACE_START,
						strlen(YAHOO_FONT_FACE_START)))
				{
					int k = 0;

					i += strlen(YAHOO_FONT_FACE_START);
					while (text && i < strlen(text) && text[i] != '"')
					{
						face[k] = text[i];
						++i;
						++k;
						face[k] = '\0';
					}
					while (text && i < strlen(text)
						&& NULL != strchr("\" \t", text[i]))
					{
						++i;
					}
				}

				if (text && '>' == text[i])
				{
					++i;
					closeFound = TRUE;
					continue;
				}

				if (!strncmp(&text[i], YAHOO_FONT_SIZE_START,
						strlen(YAHOO_FONT_SIZE_START)))
				{
					int k = 0;

					i += strlen(YAHOO_FONT_SIZE_START);
					while (text && i < strlen(text) && text[i] != '"')
					{
						size[k] = text[i];
						++i;
						++k;
						size[k] = '\0';
					}
					while (text && i < strlen(text)
						&& NULL != strchr("\" \t", text[i]))
					{
						++i;
					}
				}
			}

			if ( show_fonts ) /* should we use their font? */
			{
				font = getFont(bold, italic, underline, face, size);
			}
		}
		else if (( text[i] == 033 ) && ( text[i+1] == '[' ))
		{
			i += 2;  /* skip over the escape and [ */
			if ( text[i+1] == ';' ) {  /* we have a bold/not bold */
				if ( text[i] == '0' ) {
					bold = FALSE;
				} else {
					bold = TRUE;
				}
			    font = getFont(bold, italic, underline, face, size);
				i += 2;
			}
			if ( text[i] == 'x' ) {
				/* turn something OFF */
				i++;
				switch ( text[i] ) {
					case '1': bold = FALSE ;
							  font = getFont(bold, italic, underline, face,
							  		size);
					          break;
					case '2': italic = FALSE;
							  font = getFont(bold, italic, underline, face,
							  		size);
					          break;
					case '4': underline = FALSE;
							  font = getFont(bold, italic, underline, face,
							  		size);
					          break;
					case 'l': url = FALSE ;
							  r = or; g = og; b = ob;
							  font = getFont(bold, italic, underline, face,
							  	size);
							  break;
				}
				i += 2;
				continue;
			} else if ( text[i] == '#' ) {
				/* we found a blended color */
				char tmp_int[3];

				if ( show_blended_colors ) {
					or = r; og = g; ob = b;

					tmp_int[2] = '\0';

					tmp_int[0] = text[ i + 1 ];
					tmp_int[1] = text[ i + 2 ];
					sscanf( tmp_int, "%x", &r );

					tmp_int[0] = text[ i + 3 ];
					tmp_int[1] = text[ i + 4 ];
					sscanf( tmp_int, "%x", &g );

					tmp_int[0] = text[ i + 5 ];
					tmp_int[1] = text[ i + 6 ];
					sscanf( tmp_int, "%x", &b );
				}

			    i += 8;
			    continue;
			} else {
				tmp_color = atoi( &text[i] );
				switch ( tmp_color ) {
					case 0: if (( text[i] == 'l' ) && ( text[i+1] == 'm' )) {
							  if ( show_colors ) {
					        	url = TRUE ;
								or = r; og = g; ob = b;
								r = 0; g = 0; b = 0xff;
							  }
							  i += 2;
							  continue;
							} else if ( text[i+1] == 'm' ) {
							  italic = FALSE;
							  bold = FALSE;
							  underline = FALSE;
							  url = FALSE;
							  font = getFont(bold, italic, underline, face,
							  	size);

							  /* black */
							  or = r; og = g; ob = b;
							  r = 0; g = 0; b = 0;
							  i += 2;
							  continue;
							}
						    break;
					case 1: bold = TRUE ;
							if ( show_fonts )
								font = getFont(bold, italic, underline, face,
									size);
							i += 2;
							continue;
					        break;
					case 2: italic = TRUE;
							if ( show_fonts )
								font = getFont(bold, italic, underline, face,
									size);
							i += 2;
							continue;
					        break;
					case 4: underline = TRUE;
							if ( show_fonts )
								font = getFont(bold, italic, underline, face,
									size);
							i += 2;
							continue;
					        break;
					case 30: /* black */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0; g = 0; b = 0;
							}
							i += 3;
							break;
					case 31: /* blue */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0; g = 0; b = 0xff;
							}
							i += 3;
						    break;
					case 32: /* cyan */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0; g = 0xb2; b = 0xb2;
							}
							i += 3;
						    break;
					case 33: /* gray */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0x80; g = 0x80; b = 0x80;
							}
							i += 3;
						    break;
					case 34: /* green */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0; g = 0xc2 ; b = 0;
							}
							i += 3;
						    break;
					case 35: /* pink */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0xff; g = 0xaf; b = 0xaf;
							}
							i += 3;
						    break;
					case 36: /* purple */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0xb2; g = 0; b = 0xb2;
							}
							i += 3;
						    break;
					case 37: /* yellow */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0xff; g = 0xff ; b = 0;
							}
							i += 3;
						    break;
					case 38: /* red */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0xff ; g = 0; b = 0;
							}
							i += 3;
						    break;
					case 39: /* olive */
							if ( show_colors ) {
								or = r; og = g; ob = b;
								r = 0x54; g = 0x6b ; b = 0x50;
							}
							i += 3;
						    break;
				}
			}
			continue;
		} else if ( text[i] == 033 ) {
			i++;
		}

		if (i >= strlen(text))
		{
			break;
		}

		if ( text[i] == 033 ) {
			continue;
		}

		/* print out as much as we can */
		if (( tmp_ptr = strchr( &text[i], 033 )) != NULL ) {
			tmp_color = tmp_ptr - &text[i];
		} else {
			tmp_color = strlen( &text[i] );
		}

		if ( textbox ) {
			color->red   = r * ( 65535 / 255 );
			color->green = g * ( 65535 / 255 );
			color->blue  = b * ( 65535 / 255 );
			gdk_colormap_alloc_color(
				gtk_widget_get_colormap(GTK_WIDGET(textbox)),
				color, FALSE, TRUE);
			gtk_text_insert(GTK_TEXT(textbox), font, color,
				NULL, (gchar *) & text[i], tmp_color );
		} else {
			ct_set_color( r, g, b );
			ct_set_font( font );

			tmp_ptr = strstr( text, ":\n" );
			/* attempt to display /exec-ed stuff in fixed width font */
			if (( tmp_ptr != NULL ) &&
				( tmp_ptr != ( text + strlen( text ) - 2 ))) {
				ct_append_fixed( &text[i], tmp_color );
			} else {
				ct_append( &text[i], tmp_color );
			}
		}

		if ( capture_fp ) {
			fwrite( &text[i], 1, tmp_color, capture_fp );
		}

		i += tmp_color;
	}

	if ( ! textbox )
		ct_thaw();

	if( scroll )
		ct_scroll_to_bottom();

	if ( capture_fp )
		fflush( capture_fp );

	DBG( 22, "RETURN append_to_textbox()\n" );
}

void append_to_textbox_color(GtkWidget *win, GtkWidget *tb, char *text) {
	int old_colors = show_colors;
	int old_fonts = show_fonts;
	int old_lower = force_lowercase;

	DBG( 21, "append_to_textbox_color( %p, %p, '%s' )\n", win, tb, text );

	show_colors = 1;
	show_fonts = 1;
	force_lowercase = 0;

	/* reset things to normal */
	if ( bold )
		append_to_textbox( win, tb, YAHOO_STYLE_BOLDOFF );

	if ( italic )
		append_to_textbox( win, tb, YAHOO_STYLE_ITALICOFF );

	if ( underline )
		append_to_textbox( win, tb, YAHOO_STYLE_UNDERLINEOFF );

	if ( url )
		append_to_textbox( win, tb, YAHOO_STYLE_URLOFF );

	/* now send our text */
	append_to_textbox( win, tb, text );

	show_colors = old_colors;
	show_fonts = old_fonts;
	force_lowercase = old_lower;

	DBG( 22, "RETURN append_to_textbox_color()\n" );
}

/* ------------------------------------------------------------------------ */ 
 
/* 
 * copyright David J. Binette 
 * ALL RIGHTS RESERVED 
 */ 
/* ------------------------------------------------------------------------ */ 
/* tab stops are set to 4 */ 
/* ------------------------------------------------------------------------ */ 
/* yahmini is an small yahoo chat client as of July 8 2000 
 * written in portable C for win32 and linux 
 * Its main grace is its small size partially attributable 
 * to consideration of the segment boundarys created in win32 applications. 
 * yahmini has no global volatile variables ,hence 
 * there is only a code segment and a const data segment. 
 * *your* C runtime however may add a whole pile of crap 
 * making it significantly larger. 
 * 
 */ 
 
 
/* ------------------------------------------------------------------------ */ 
 
/* the http request to get cookie ,with username ,password and host */ 
static const char xHTTPGet_argU_argP_argH[] = { 
"GET /config/ncclogin?.src=bl&login=%s&passwd=%s&n=1 HTTP/1.0\r\n" 
"Host: %s\r\n" 
"Accept: */*\r\n" 
"User-Agent: Mozilla/4.0\r\n" 
"\r\n" 
}; 
static const char HTTPGet_argU_argP_argH[] = { 
"GET %s/config/login?login=%s&passwd=%s&tries=1&.src=chat&.persistent=Y&.chkP=Y HTTP/1.0\r\n" 
"Host: %s\r\n" 
"Accept: */*\r\n" 
"User-Agent: Mozilla/4.0\r\n" 
"\r\n" 
}; 
 
/* ------------------------------------------------------------------------ */ 
 
static void showSockError(SESSION *sess ,const char * arg1 ,const char *arg2 
                         ,const char *arg3 ,int errcode) 
{ 
	char buf[4096];

    if (!sess->quit) 
    {
        sprintf(buf 
                ,"%s%s %s failed %s, error %d\n" ,"*** " 
                ,arg1 ,arg2 ,arg3 ,errcode); 
        append_to_textbox( chat_window, NULL, buf );
    }
} 
 
/* ------------------------------------------------------------------------ */ 
/* 
 * wait for up to 250mS for a packet 
 * RETURNS 
 *      <0 error ,socket is closed 
 *      =0 timeout ,no data available 
 *      >0 data waiting 
 */ 
/* ------------------------------------------------------------------------ */ 
static int sockDataAvailable(SOCKET sock) 
{ 
    int rv; 
    struct timeval  tv; 
    fd_set          fds; 
    tv.tv_sec       = 0; 
    tv.tv_usec      = 100000; 
    FD_ZERO(&fds); 
    FD_SET(sock ,&fds); 
 
    rv = select(sock+1 ,&fds ,NULL ,NULL ,&tv); 
 
    if (rv > 0) 
    {   char x[2]; 
        if (recv(sock ,x ,1 ,MSG_PEEK) < 1) 
            rv = -1; 
    } 
 
    return rv; 
} 
/* ------------------------------------------------------------------------ */ 
 
static int selectRecv(SESSION *sess ,char *buf ,int siz ,int flags) 
{ 
	while(1)
    { 
		waitpid( -1, NULL, WNOHANG );

		if ( prof_text )
			display_profile();

        if (sockDataAvailable(sess->sock)) 
            return recv(sess->sock ,buf ,siz ,flags); 
    } 

	return FALSE;
} 
 
/* ------------------------------------------------------------------------ */ 
/* General purpose socket read 
 * sock         socket 
 * buf          pointer to data buffer 
 * siz          size of buffer 
 * info         text for error message 
 * 
 * RETURN 
 *        >= 0 number bytes received 
 *        <  0 on error 
 */ 
static int myRecv(SESSION *sess ,char *buf ,int siz ,const char *info) 
{ 
    int cnt = 0; 
    while (siz) 
    {   int b = selectRecv(sess ,buf ,siz ,0); 
        if (b == -1) 
        {
			showSockError(sess ,info,"incomplete" ,"recv()" ,SockErrNum()); 
			on_disconnect_activate( NULL, NULL );
            return -1; 
        } 
        if (b == 0) 
            break; 
        cnt += b; 
        buf += b; 
        siz -= b; 
    } 
    return cnt; 
} 
 
/* ------------------------------------------------------------------------ */ 
/* read exactly n bytes 
 * Function will return TRUE if success 
 */ 
 
BOOL recvN(SESSION *sess ,char *buf ,int cnt ,const char *info) 
{ 
    while (cnt) 
    {   int b = myRecv(sess ,buf ,cnt ,info); 
        if (b < 1) 
        {
			/*
			showSockError(sess ,info ,"incomplete" ,"recv()" ,SockErrNum()); 
			*/
            break; 
        } 
        buf += b; 
        cnt -= b; 
    } 
    return cnt==0; 
} 
 
/* ------------------------------------------------------------------------ */ 
/* send exactly n bytes 
 * Function will return TRUE if success 
 */ 
 
BOOL sendN(SESSION *sess ,const u_char *buf ,int cnt ,const char *info) 
{ 
    while (cnt) 
    {   int b = send(sess->sock ,(char*)buf ,cnt ,0); 
        if (b < 1) 
        {   showSockError(sess ,info ,"incomplete" ,"send()" ,SockErrNum()); 
            break; 
        } 
        buf += b; 
        cnt -= b; 
    } 
    return cnt==0; 
} 
 
/* ------------------------------------------------------------------------ */ 
/* A function to resolve and connect to a TCP/IP port 
 * return INVALID_SOCKET on error 
 * otherwise return a valid socket handle 
 */ 
static SOCKET setupConnection(SESSION *sess 
     ,const char *host ,const u_short port ,const int proto ,const char *info) 
{ 
    SOCKET sock; 
    struct sockaddr_in serv_addr; 
	char buf[4096];
 
    sprintf(buf ,"Finding %s..." ,host); 

	gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid, buf );
	process_gtk_events();

    memset(&serv_addr ,0 ,sizeof(serv_addr)); 
    serv_addr.sin_addr.s_addr = inet_addr(host); 
 
    if (serv_addr.sin_addr.s_addr == (u_int)INADDR_NONE) 
    {   /* resolve hostname to ipaddr */ 
        struct hostent *hostInfo = gethostbyname(host); 
 
        if (!hostInfo || (hostInfo->h_addrtype != AF_INET)) 
        {   showSockError(sess ,info ,host ,"gethostbyname()" ,SockErrNum()); 
            return INVALID_SOCKET; 
        } 
 
        memset(&serv_addr ,0  ,sizeof(serv_addr)); 
        memmove((void*)&serv_addr.sin_addr.s_addr 
                ,hostInfo->h_addr 
                ,hostInfo->h_length); 
    } 

	gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
	process_gtk_events();
    sprintf(buf ,"Connecting to %s..." ,host); 
	gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid, buf );
	process_gtk_events();
 
    serv_addr.sin_family = AF_INET; 
    serv_addr.sin_port   = htons(port); 
 
    sock = socket(AF_INET ,SOCK_STREAM ,proto); 
    if (sock == INVALID_SOCKET) {
        showSockError(sess ,info ,host ,"socket()" ,SockErrNum()); 
	} else if (connect(sock ,(struct sockaddr*)&serv_addr ,sizeof(serv_addr)) == -1) 
    {
		showSockError(sess ,info ,host ,"connect()" ,SockErrNum()); 
        closesocket(sock); 
        sock = INVALID_SOCKET; 
		gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
		process_gtk_events();
    } 

	gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
	process_gtk_events();
 
    return sock; 
} 

/* ------------------------------------------------------------------------ */ 
/* login to yahoo authentication server and get cookie 
 * return TRUE if success 
 */ 
 
static BOOL subLoginToYahoo(SESSION *sess ,char *buf ,int siz) 
{ 
/*
    int   n; 
    char *p; 
    char *hostname = SERVERAUTHHOST; 
	char proxy_str[1024];
*/
 
#if USETINYCRT 
    char *np; 
#define strnext(a ,b) strtoken(a ,b ,&np) 
#else 
#define strnext(a ,b) strtok(a ,b) 
#endif 
 
    sess->cook[0] = EOS; 

	gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
		"Requesting Login Key..." );
	process_gtk_events();

	if ( use_proxy ) {
	    sess->sock = setupConnection(sess ,proxy_host ,proxy_port ,IPPROTO_TCP ,"Login"); 
	} else {
	    sess->sock = setupConnection( sess, sess->host, YMSG9_CHAT_PORT,
			IPPROTO_TCP, "Login"); 
	}

#ifndef OS_WINDOWS
	io_callback_tag = gdk_input_add( sess->sock, GDK_INPUT_READ,
		io_callback, (gpointer) 0);
#endif

	ymsg9_request_key();
	gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
	return TRUE;

#ifdef YCHT
    if (sess->sock != INVALID_SOCKET) 
    { 
		gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );

		if ( use_proxy ) {
			sprintf( proxy_str, "http://%s:%d", hostname, SERVERAUTHPORT );
			n = sprintf(buf ,HTTPGet_argU_argP_argH ,
				proxy_str, sess->user ,sess->pass , hostname); 
		} else {
			n = sprintf(buf ,HTTPGet_argU_argP_argH ,
				"", sess->user ,sess->pass ,hostname); 
		}
		gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
			"Sending Login Sequence..." );
		process_gtk_events();
        n = sendN(sess ,(const u_char*)buf ,n ,"Login"); 
		gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
 
        if (n) {
			gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
				"Waiting for Login response..." );
			process_gtk_events();
            n = myRecv(sess ,buf ,siz-1 ,"Login"); 
			gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
		}
 
        closesocket(sess->sock); 
        sess->sock = INVALID_SOCKET; 
 
        if (n < 1) 
        {
			append_to_textbox( chat_window, NULL, "*** no login data\n" );
		}
        else 
        {   buf[n] = EOS; 
            /* we have to be careful what we output 
             * because some terminals allow keys to be remapped 
             * we dont want to every allow unstripped data to be 
             * sent to the terminal 
             */ 
            for(p=strnext(buf ,"\r\n"); p; p=strnext(NULL ,"\r\n")) 
            { 
                if (!strcmp(p ,"ERROR: Invalid NCC Login")) 
                {   sess->cook[0] = EOS; 
					append_to_textbox( chat_window, NULL, p );
                    break; 
                } 
 
                if (!strncmp(p ,"Set-Cookie:" ,11)) 
                {   char *q = strchr(p ,';'); 

                    if (q) 
                        *q = EOS; 
                    q = p+11; 
 
                    q = skip_whitespace(q); 
 
                    if (*q) 
                    {   if ((strlen(sess->cook)+strlen(q)+3) >= CHATCOOKSIZE) 
                    	{
                            append_to_textbox( chat_window, NULL,
                            	"*** Bad cookie\n" );
                        }
                        else 
                        {   if (sess->cook[0]) 
                                strcat(sess->cook ,"; "); 
							if ( strstr( q, "T=" ) || strstr( q, "Y=" )) {
                            	strcat(sess->cook ,q); 
							}
                        } 
                    } 
                } 
            } 
        } 
    } 
#endif /* YCHT */
	gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
    return sess->cook[0] ? TRUE : FALSE; 
} 
 
/* wrap the call to handle the malloc()/free() 
 * I dont like big stack frames 
 * return TRUE if success 
 */ 
BOOL LoginToYahoo(SESSION *sess) 
{ 
    BOOL r    = FALSE; 
    char *buf = malloc(65534); 
 
    if (buf) 
    {   r    = subLoginToYahoo(sess ,buf ,65533); 
        free(buf); 
    } 
    else 
    {
		append_to_textbox( chat_window, NULL, "no mem\n" );
    }
    return r; 
} 
 
void subst_args( char *str, char *user )
{
	char tmp[4096];
	char *from, *to;

	DBG( 21, "subst_args( '%s', '%s' )\n", str, user );

	from = str;
	to = tmp;

	if ( ! strchr( str, '%' )) return;

	while( *from ) {
		if (( *from == '%' ) && ( *(from+1) == 'N' )) {
			strcpy( to, YAHOO_COLOR_RED );
			to += strlen( YAHOO_COLOR_RED );
			strcpy( to, user );
			to += strlen( user );
			strcpy( to, YAHOO_COLOR_PURPLE );
			to += strlen( YAHOO_COLOR_PURPLE );
			from += 2;
		} else {
			*(to++) = *(from++);
		}
	}
	*to = '\0';

	strcpy( str, tmp );

	DBG( 22, "RETURN subst_args() == '%s'\n", str );

	return;
}

int check_alias(SESSION *sess, char *emote, char *args )
{
	gchar *listitem[3];
	gchar *emotion;
	gchar *em_everyone;
	gchar *em_user;
	u_char  tmp[4096] = "";
	u_char  tmp2[4096] = "";
	u_char  tmp_em[21];
	int i, x;
	int found = -1;

	DBG( 21, "check_alias( %p, '%s', '%s' )\n", sess, emote, args );

	listitem[0] = emotion;
	listitem[1] = em_everyone;
	listitem[2] = em_user;

	for( i = 0; i < strlen( emote ); i++ ) {
		emote[i] = tolower( emote[i] );
	}

	i = 0;
	while(1) {
		if ( !gtk_clist_get_text(GTK_CLIST(chat_emotions), i, 0, &emotion ))
			break;

		strcpy( tmp_em, emotion );
		for( x = 0; x < strlen( tmp_em ); x++ ) {
			tmp_em[x] = tolower( tmp_em[x] );
		}

		if ( ! strcmp( emote, tmp_em )) {
			emotion = tmp_em;
			found = i;
			break;
		}

		i++;
	}

	if ( found >= 0 ) {
		if ( args[0] ) {
			gtk_clist_get_text(GTK_CLIST(chat_emotions), found, 2,
				&em_user );

			strcpy( tmp2, em_user );

			if ( strchr( tmp2, '$' )) {
				strcpy( tmp2, replace_args( tmp2, args ));
				args[0] = '\0';
			}

			em_user = tmp2;
			if (( em_user[0] == '/' ) ||
				( em_user[0] == '!' )) {
				strcpy( tmp, em_user );
				if ( args[0] ) {
					strcat( tmp, " " );
					strcat( tmp, args );
				}
				chatOut( sess, tmp );
				DBG( 22, "RETURN check_alias() == TRUE\n" );

				return TRUE;
			}

			strcpy( tmp, YAHOO_COLOR_PURPLE );
			strcat( tmp, em_user );

		} else {
			gtk_clist_get_text(GTK_CLIST(chat_emotions), found, 1,
				&em_everyone );

			if (( em_everyone[0] == '/' ) ||
				( em_everyone[0] == '!' )) {
				chatOut( sess, em_everyone );
				DBG( 22, "RETURN check_alias() == TRUE\n" );

				return TRUE;
			}

			strcat( tmp, YAHOO_COLOR_PURPLE );
			strcpy( tmp, em_everyone );
		}

		subst_escs( tmp );

		if ( my_status )
			cmd_away( "0" );

		if( send_avatar ) {
			strcat( tmp, AVATAR_START );
			strcat( tmp, send_avatar );
			strcat( tmp, AVATAR_SEND_END );
			display_avatar( sess->user, send_avatar );
		}

		ymsg9_emote( tmp );

		DBG( 22, "RETURN check_alias() == TRUE\n" );

		return( TRUE );
	} else {
		DBG( 22, "RETURN check_alias() == FALSE\n" );

		return( FALSE );
	}
}

/* ------------------------------------------------------------------------ */ 
/* send a chat command 
 * Note: it modifies the string in place 
 * return FALSE if we should quit 
 */ 
BOOL chatOut(SESSION *sess ,char *cmd) 
{ 
	BOOL r    = TRUE; 
	char *txt = ""; 
	char *args;
	char buf[1025] = "";
	regex_t smiley_regex;

	DBG( 21, "chatOut( %p, '%s' )\n", sess, cmd );

	cmd = skip_whitespace( cmd );

	auto_away_time = time(NULL) + ( auto_away * 60 );

	switch(cmd[0]) { 
		case EOS: 
			return( TRUE );
			break; 
		case ':':
			/* differentiate between smileys and emotes */
			regcomp( &smiley_regex, "^:[)({}|<>DP]", REG_NOSUB );
			
			if ( ! regexec( &smiley_regex, cmd, 0, NULL, 0 )) {
				/* it's a smiley */
				regfree( &smiley_regex );
				txt = cmd;
				subst_escs( txt );
			} else {
				/* it's not */
				regfree( &smiley_regex );
				sprintf( buf, "/emote %s", cmd+1 );
				try_command( buf );
				return( TRUE );
			}
			break;
		case '!':
			if ( cmd[1] == '!' ) {
				sprintf( buf, "/execlocal %s", cmd+2 );
			} else {
				sprintf( buf, "/exec %s", cmd+1 );
			}
			try_command( buf );
			return( TRUE );
			break;

		case '/': 
			if ( try_command( cmd ))
				return( TRUE );

			args = find_whitespace( cmd + 1);
			if ( *args ) {
				*args = '\0';
				args++;
				args = skip_whitespace( args );
			}

			if ( check_alias( sess, cmd + 1, args ))
				return( TRUE );
        	break; 
 
		default: 
			if (( cmd[0] == '\\' ) &&
				(( cmd[1] == '/' ) ||
				 ( cmd[1] == ':' ) ||
				 ( cmd[1] == '!' ))) {
				txt = cmd +1;
			} else {
				txt = cmd; 
			}

			subst_escs( txt );
			break;
    } 

	if ( use_color ) {
		if ( ! strcmp( use_color, "black" )) {
			strcat( buf, YAHOO_COLOR_BLACK );
		} else if ( ! strcmp( use_color, "blue" )) {
			strcat( buf, YAHOO_COLOR_BLUE );
		} else if ( ! strcmp( use_color, "cyan" )) {
			strcat( buf, YAHOO_COLOR_CYAN );
		} else if ( ! strcmp( use_color, "gray" )) {
			strcat( buf, YAHOO_COLOR_GRAY );
		} else if ( ! strcmp( use_color, "green" )) {
			strcat( buf, YAHOO_COLOR_GREEN );
		} else if ( ! strcmp( use_color, "pink" )) {
			strcat( buf, YAHOO_COLOR_PINK );
		} else if ( ! strcmp( use_color, "purple" )) {
			strcat( buf, YAHOO_COLOR_PURPLE );
		} else if ( ! strcmp( use_color, "yellow" )) {
			strcat( buf, "\033[#ffff00m" );
		} else if ( ! strcmp( use_color, "orange" )) {
			strcat( buf, "\033[#ffc800m" );
		} else if ( ! strcmp( use_color, "red" )) {
			strcat( buf, YAHOO_COLOR_RED );
		} else if ( ! strcmp( use_color, "custom" )) {
			sprintf( buf, "\033[#%2x%2x%2xm",
				custom_color_red, custom_color_green, custom_color_blue );
		}
	}

	if ( use_bold )
		strcat( buf, YAHOO_STYLE_BOLDON );

	if ( use_italics )
		strcat( buf, YAHOO_STYLE_ITALICON );

	if ( use_underline )
		strcat( buf, YAHOO_STYLE_UNDERLINEON );

	if ( filter_command ) {
		strcat( buf, filter_text( txt ));
	} else {
		strcat( buf, txt );
	}

    if ( *buf ) {
		if( send_avatar ) {
			strcat( buf, AVATAR_START );
			strcat( buf, send_avatar );
			strcat( buf, AVATAR_SEND_END );
			display_avatar( sess->user, send_avatar );
		}

		ymsg9_comment( buf );
		r = 1;
	}
 
	DBG( 22, "RETURN chatOut() == %d\n", r );

    return r; 
} 
 
char ymsg_fields[300][1024];

/* It would not be too hard to modify this function to return a -1 if the */
/* number of fields in the input string was not equal to the number of    */
/* variables passed that the results are placed in.  Right now, it will   */
/* process until it runs out of variables or fields, and returns the	  */
/* number of items it was able to process.				  */

int split( char *str, char *connector )
{
	char	*last;
	char	*orig;
	char	*ptr;
	char	tmp_str[ 2048 ];
	int	result = 0;
	int i = 0;

	strcpy( tmp_str, str );
	str = tmp_str;

	last = orig = str;

	while (( ptr = strstr( str, connector ))) 
	{
		*ptr = '\0';
		strcpy( ymsg_fields[result], last );
		str = last = ptr + strlen( connector );
		result++;
	}

	if ( last != orig ) {
		strcpy( ymsg_fields[result], last );
	} else {
		result--;
	}

	strcpy( ymsg_fields[result], "__END__" );
	strcpy( ymsg_fields[result+1], "__END__" );

	i = 0;
	while( strcmp( ymsg_fields[i], "__END__" )) {
		if (( debug_packets ) &&
			( strcmp( ymsg_fields[i], "117" ))) {
			fprintf( stderr, "%-3s = '%s'\n",
				ymsg_fields[i], ymsg_fields[i+1] );
			fflush( stderr );
		}
		i += 2;
	}

	return( result + 1 );
}


int join( char *str, char *connector, ... )
{
	char	*temp;
	int	result = 0;
	va_list	arg_ptr;


#ifdef UNUSUAL_VA_START
	va_start( arg_ptr );
#else
	va_start( arg_ptr, connector );
#endif

	str[ 0 ] = 0;

	while (( temp = va_arg( arg_ptr, char *)) != NULL )
	{
		strcat( str, temp );
		strcat( str, connector );
		result++;
	}

	temp = str + strlen( str ) - strlen( connector );
	*temp = 0;

	va_end( arg_ptr );

	return( result );
}

char *ymsg9_field_p( char *key ) {
	int i = 0;
	static char result[2048];

	result[0] = '\0';
	while( strcmp( ymsg_fields[i], "__END__" )) {
		if ( ! strcmp( ymsg_fields[i], key )) {
			if ( result[0] ) {
				strcat( result, "|" );
			}
			strcat( result, ymsg_fields[i+1] );
		}
		i += 2;
	}
	return( result );
}

char *ymsg9_field( char *key ) {
	int i = 0;
	static char result[2048];

	result[0] = '\0';
	while( strcmp( ymsg_fields[i], "__END__" )) {
		if ( ! strcmp( ymsg_fields[i], key )) {
			if ( result[0] ) {
				strcat( result, "," );
			}
			strcat( result, ymsg_fields[i+1] );
		}
		i += 2;
	}
	return( result );
}

void chatter_list_populate( char *list ) {
	char *user;
	char *end;
	int last = 0;
	int row;
	int users = 0;
	char buf[256];
	gchar *listitem[3];
	GtkWidget *tmp_widget = NULL;
	GtkStyle *user_style = NULL;

#ifdef SHOW_STATUS_PIXMAPS
	GdkPixmap *st_pixmap = NULL;
	GdkBitmap *st_mask = NULL;
#else
	char status[5] = "";
#endif

#ifdef CASE_INSENSITIVE_USER_LIST
	char *temp_ptr;
#endif



	gtk_clist_freeze(GTK_CLIST(chat_users));
	gtk_clist_clear(GTK_CLIST(chat_users));

	user_list = gyach_g_list_free( user_list );
	full_list = gyach_g_list_free( full_list );
	left_list = gyach_g_list_free( left_list );
	tab_user = NULL;

	if ( show_avatars ) {
		gtk_clist_set_column_width( GTK_CLIST(chat_users), 0, 20 );
		gtk_clist_set_row_height( GTK_CLIST(chat_users), 0 );
	}

	tmp_widget = lookup_widget( chat_window, "users" );
	user_style = gtk_widget_get_style( tmp_widget );

	user = list;
	while( user ) {
		end = strchr( user, ',' );
		if ( end ) {
			*end = '\0';
		} else {
			last = 1;
		}

		listitem[0] = user;
		listitem[1] = user;
		listitem[2] = user;
#ifdef CASE_INSENSITIVE_USER_LIST
		tmp_ptr = user;
		while( *tmp_ptr ) {
			*tmp_ptr = tolower( *tmp_ptr );
			tmp_ptr++;
		}
#endif

#ifdef SHOW_STATUS_PIXMAPS
		if ( ignore_check( user )) {
			st_pixmap = gdk_pixmap_create_from_xpm_d(
				tmp_widget->window, &st_mask,
				&user_style->bg[GTK_STATE_NORMAL],
				(gchar **)pixmap_status_ignore );
		} else {
			if ( mute_check( user )) {
				st_pixmap = gdk_pixmap_create_from_xpm_d(
					tmp_widget->window, &st_mask,
					&user_style->bg[GTK_STATE_NORMAL],
					(gchar **)pixmap_status_mute );
			} else {
				st_pixmap = gdk_pixmap_create_from_xpm_d(
					tmp_widget->window, &st_mask,
					&user_style->bg[GTK_STATE_NORMAL],
					(gchar **)pixmap_status_here );
			}
		}
#else
		if ( ignore_check( user )) {
			strcpy( status, "IG" );
		} else {
			if ( mute_check( user )) {
				strcpy( status, "MU" );
			} else {
				strcpy( status, "" );
			}
		}
#endif
			
		row = find_user_row( user );
		if ( row < 0 ) {
			users++;
			user_add( user );

			row = gtk_clist_insert(GTK_CLIST(chat_users), 0, listitem );
			gtk_clist_set_text(GTK_CLIST(chat_users), row, 1, user );
			gtk_clist_set_text(GTK_CLIST(chat_users), row, 2, "000:" );
#ifdef SHOW_STATUS_PIXMAPS
			gtk_clist_set_pixmap(GTK_CLIST(chat_users), row, 0,
				st_pixmap, st_mask );
			gdk_pixmap_unref( st_pixmap );
			gdk_bitmap_unref( st_mask );
#else
			gtk_clist_set_text(GTK_CLIST(chat_users), row, 0, status );
#endif
		}

		if ( last ) {
			user = NULL;
		} else {
			user = end + 1;
		}
	}


	gtk_clist_sort(GTK_CLIST(chat_users));
	gtk_clist_thaw(GTK_CLIST(chat_users));

	gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
    sprintf(buf ,"Current Room: %s  (%d users)", sess->room, users ); 
	gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid, buf );
}

void chatter_list_add( char *user ) {
	int row;
	int iggy;
	int users;
	char buf[256];
	gchar *listitem[3];
	GtkWidget *tmp_widget = NULL;
	GtkStyle *user_style = NULL;

#ifdef SHOW_STATUS_PIXMAPS
	GdkPixmap *st_pixmap = NULL;
	GdkBitmap *st_mask = NULL;
#else
	char status[5] = "";
#endif

#ifdef CASE_INSENSITIVE_USER_LIST
	char *temp_ptr;
#endif


	tmp_widget = lookup_widget( chat_window, "users" );
	user_style = gtk_widget_get_style( tmp_widget );

	row = find_user_row( user );
	if ( row != -1 ) {
		/* user is already in our list */
		return;
	}

	/* entering */
	listitem[0] = user;
	listitem[1] = user;
	listitem[2] = user;
#ifdef CASE_INSENSITIVE_USER_LIST
	ptr = user;
	while( *ptr ) {
		*ptr = tolower( *ptr );
		ptr++;
	}
#endif

#ifdef SHOW_STATUS_PIXMAPS
	if (( iggy = ignore_check( user )) != 0 ) {
		st_pixmap = gdk_pixmap_create_from_xpm_d(
			tmp_widget->window, &st_mask,
			&user_style->bg[GTK_STATE_NORMAL],
			(gchar **)pixmap_status_ignore );
	} else {
		if (( iggy = mute_check( user )) != 0 ) {
			st_pixmap = gdk_pixmap_create_from_xpm_d(
				tmp_widget->window, &st_mask,
				&user_style->bg[GTK_STATE_NORMAL],
				(gchar **)pixmap_status_mute );
		} else {
			st_pixmap = gdk_pixmap_create_from_xpm_d(
				tmp_widget->window, &st_mask,
				&user_style->bg[GTK_STATE_NORMAL],
				(gchar **)pixmap_status_here );
		}
	}
#else
	if (( iggy = ignore_check( user )) != 0 ) {
		strcpy( status, "IG" );
	} else {
		if (( iggy = mute_check( user )) != 0 ) {
			strcpy( status, "MU" );
		} else {
			strcpy( status, "" );
		}
	}
#endif

	user_add( user );
	gtk_clist_freeze(GTK_CLIST(chat_users));
	row = gtk_clist_append(GTK_CLIST(chat_users), listitem );
	gtk_clist_set_text(GTK_CLIST(chat_users), row, 1, user );
	gtk_clist_set_text(GTK_CLIST(chat_users), row, 2, "000:" );

#ifdef SHOW_STATUS_PIXMAPS
	gtk_clist_set_pixmap(GTK_CLIST(chat_users), row, 0,
		st_pixmap, st_mask );
	gdk_pixmap_unref( st_pixmap );
	gdk_bitmap_unref( st_mask );
#else
	gtk_clist_set_text(GTK_CLIST(chat_users), row, 0, status );
#endif

	gtk_clist_sort(GTK_CLIST(chat_users));
	gtk_clist_thaw(GTK_CLIST(chat_users));

	users = GTK_CLIST(chat_users)->rows;
	gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
    sprintf(buf ,"Current Room: %s  (%d users)", sess->room, users ); 
	gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid, buf );
}

void chatter_list_remove( char *user ) {
	int row;
	int users;
	char buf[256];

	row = find_user_row( user );

	if ( row >= 0 ) {
		user_remove( user );
		gtk_clist_remove( GTK_CLIST(chat_users), row );
	}

	if (( follow_user ) &&
		( ! strcmp( user, follow_user ))) {
		sprintf( buf, "/goto %s", follow_user );
		try_command( buf );
	}

	users = GTK_CLIST(chat_users)->rows;
	gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
    sprintf(buf ,"Current Room: %s  (%d users)", sess->room, users ); 
	gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid, buf );
}

void chatter_list_status( char *user, gchar **status_image, char *status ) {
	int row;
#ifdef SHOW_STATUS_PIXMAPS
	GtkWidget *tmp_widget = NULL;
	GtkStyle *user_style = NULL;
	GdkPixmap *st_pixmap = NULL;
	GdkBitmap *st_mask = NULL;
#endif

	row = find_user_row( user );
#ifdef SHOW_STATUS_PIXMAPS
	tmp_widget = lookup_widget( chat_window, "users" );
	user_style = gtk_widget_get_style( tmp_widget );
	st_pixmap = gdk_pixmap_create_from_xpm_d(
		tmp_widget->window, &st_mask,
		&user_style->bg[GTK_STATE_NORMAL],
		(gchar **)status_image );
	gtk_clist_set_pixmap(GTK_CLIST(chat_users), row, 0,
		st_pixmap, st_mask );
	gdk_pixmap_unref( st_pixmap );
	gdk_bitmap_unref( st_mask );
#else
	gtk_clist_set_text(GTK_CLIST(chat_users), row, 0, status );
#endif
}

void showYahooPacket(SESSION *sess) {
	static char buf[1024];  /* these are static so they aren't on the stack */
	static char tmp[2048];
	static char tmp2[2048];
	static char tmp3[2048];
	char *lf;
	char *ptr;
	char *src;
	char *end;
	int lf_count = 0;
	int last;
	time_t time_now;
	int hours;
	int minutes;
	int seconds;
	struct tm *tm_now;
	GList *pm_lptr;
	PM_SESSION *pm_sess;
	int display;
	int row;
	int accept_pm;
	int count;
	time_t exp_time;

	if ( sess->yin.size ) {
		split( sess->yin.data, "\xC0\x80" );
	} else {
		strcpy( ymsg_fields[0], "__END__" );
		strcpy( ymsg_fields[1], "__END__" );
	}

	switch( sess->yin.type ) {
		case YMSG9_GET_KEY:
				connect_time = time(NULL);

				strcpy( tmp, ymsg9_field( "94" ));
				ymsg9_login( tmp );

				gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
				gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
					"Logging in to chat server..." );

				sprintf(buf ,"Chat started, type /help for help\n"); 
				append_to_textbox( chat_window, NULL, buf );

				/* reset cookies */
				sess->cook[0] = '\0';
				break;

		case YMSG9_COOKIE:
				/* set online if we got the right cookie packet */
				if ( strcmp( ymsg9_field( "3" ), "" )) {
					gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
					gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
						"Setting status to Online..." );

					ymsg9_online();
				}

				/* store the cookies */
				if ( sess->cook[0] ) {
					strcat( sess->cook, ";" );
					strcat( sess->cook, ymsg9_field_p( "59" ));
				} else {
					strcpy( sess->cook, ymsg9_field_p( "59" ));
				}
				ptr = sess->cook;
				ptr = strchr( ptr, '	' ); /* a tab */
				while( ptr ) {
					*ptr = '=';
					ptr = strchr( ptr, '|' );

					if ( ptr ) {
						*ptr = ';';
						ptr = strchr( ptr, '	' ); /* a tab */
					}
				}

				/* set our buddy list, 88 might be ignore list */
				if ( strcmp( ymsg9_field( "87" ), "" )) {
					populate_friend_list( ymsg9_field( "87" ));
				}
				/* merge yahoo ignore list into local list */
				if ( strcmp( ymsg9_field( "88" ), "" )) {
					strcpy( tmp, ymsg9_field( "88" ));
					ptr = tmp;
					while( ptr ) {
						end = strchr( ptr, ',' );
						if ( end )
							*end = '\0';

						if ( ! ignore_check( ptr ))
							ignore_toggle( ptr );

						if ( end ) {
							ptr = end + 1;
						} else {
							ptr = NULL;
						}
					}
				}
				break;

		case YMSG9_ONLINE:
				ymsg9_join();

				gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
				sprintf( buf, "Current Room: %s", sess->room );
				gtk_statusbar_push( GTK_STATUSBAR(chat_status),
					st_cid, buf );

				/* set the return value for the program */ 
				sess->rval = 0;
				break;

		case YMSG9_MAIL:
				sprintf( buf,
					"You have %s%s%s email(s).  http://mail.yahoo.com\n",
					YAHOO_COLOR_RED, ymsg9_field( "9" ), YAHOO_COLOR_BLACK );
				append_to_textbox( chat_window, NULL, buf );
				break;

		case YMSG9_COMMENT: 
				/* user comment/emote/thought */
				strcpy( tmp, ymsg9_field( "109" ));
				strcpy( tmp2, ymsg9_field( "117" ));

				/* set auto-away if needed */
				/* should put this in the event loop somehow */
				if (( ! my_status ) &&
					( auto_away ) &&
					( time(NULL) > auto_away_time )) {
					cmd_away( "11" );
				}

				if ( ignore_check( tmp )) {
					return;
				}

				if ( mute_check( tmp )) {
					return;
				}

				if (( ptr = auto_ignore_check( tmp2 )) != NULL ) {
					if ( ! ignore_check( tmp )) {
						ignore_toggle( tmp );
					} else {
						return;
					}

					sprintf( buf, "Gyach - auto-ignored %s%s%s, regex '%s'\n",
						YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK, ptr );
					append_to_textbox( chat_window, NULL, buf );

					if ( ! disp_auto_ignored ) {
						return;
					}
				} 

				/* keep track of last thing the user said */
				row = find_user_row( tmp );
				if ( row >= 0 ) {
					gtk_clist_get_text(GTK_CLIST(chat_users), row, 2, &ptr );
					if (( ptr ) &&
						( ! strcmp( ptr + 4, tmp2 ))) {
						/* user said same thing again */
						count = atoi( ptr ) + 1;
						sprintf( buf, "%03d:%s", count, tmp2 );
						gtk_clist_set_text(GTK_CLIST(chat_users), row, 2, buf );

						if (( suppress_mult ) &&
							( count > 1 )) {
							if (( ignore_on_mults ) &&
								( count > ignore_on_mults ) &&
								( ! ignore_check( tmp ))) {
								ignore_toggle( tmp );
								sprintf( buf,
									"Gyach - auto-ignored %s%s%s, same post "
									"%d times\n", YAHOO_COLOR_RED, tmp,
									YAHOO_COLOR_BLACK, count );
								append_to_textbox( chat_window, NULL, buf );
							}
							if (( mute_on_mults ) &&
								( count > mute_on_mults ) &&
								( ! mute_check( tmp ))) {
								mute_toggle( tmp );
								sprintf( buf,
									"Gyach - auto-muted %s%s%s, same post "
									"%d times\n", YAHOO_COLOR_RED, tmp,
									YAHOO_COLOR_BLACK, count );
								append_to_textbox( chat_window, NULL, buf );
							}
							return;
						}
					} else {
						/* user said something new */
						sprintf( buf, "001:%s", tmp2 );
						gtk_clist_set_text(GTK_CLIST(chat_users), row, 2, buf );
					}
				}

				/* check for lots of linefeeds */
				if ( limit_lfs ) {
					lf = strchr( tmp2, '\n' );
					while( lf ) {
						lf_count++;
						lf = strchr( lf + 1, '\n' );
					}
					if ( lf_count > limit_lfs ) {
						lf = strchr( tmp2, '\n' );
						while( lf ) {
							*lf = '|';
							lf = strchr( lf + 1, '\n' );
						}
					}
				}

				/* need to make fetch_avatar windows-safe */
#ifndef OS_WINDOWS
				if ( show_avatars ) {
					/* check for an avatar and update if needed */
					if (( ptr = strstr( tmp2, AVATAR_START )) != NULL ) {
						/* avatar exists so try to display if unset */
						ptr += strlen( AVATAR_START );
						end = strstr( ptr, AVATAR_END );
						if ( ! end ) {
							end = strchr( ptr, '>' );
						}
						if ( end ) {
							*end = '\0';
							display_avatar( tmp, ptr );
						}
					}
				}
#endif

				strcpy( tmp2, ymsg9_field( "117" ));
				if ( chat_timestamp ) {
					time_now = time( NULL );
					tm_now = localtime( &time_now );
					sprintf( buf, "%s[%d/%d/%d %02d:%02d] ",
						YAHOO_COLOR_BLACK, tm_now->tm_mon + 1, tm_now->tm_mday,
						tm_now->tm_year + 1900, tm_now->tm_hour,
						tm_now->tm_min );
					append_to_textbox_color( chat_window, NULL, buf );
				}

				if ( strcmp( tmp, sess->user )) {
					sprintf( buf, "%s%s%s%s%s",
						YAHOO_STYLE_BOLDON, YAHOO_COLOR_RED, tmp,
						YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF );
				} else {
					/* this never really happens, we don't recv our comments */
					sprintf( buf, "%s%s%s%s%s",
						YAHOO_STYLE_BOLDON, YAHOO_COLOR_BLUE, tmp,
						YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF );
				}

				if ( ! strcmp( ymsg9_field( "124"), "1" ))
					strcat( buf, ": " );
				else
					strcat( buf, " " );

				append_to_textbox_color( chat_window, NULL, buf );

				if ( ! strcmp( ymsg9_field( "124"), "1" )) {
					sprintf( buf, "%s%s%s\n",
						YAHOO_COLOR_BLACK, tmp2, ATTR_RESET );
				} else {
					sprintf( buf, "%s%s%s\n",
						YAHOO_COLOR_PURPLE, tmp2, ATTR_RESET );

					/* check for status change */
					if ( ! strncasecmp( tmp2, "is away ", 8 )) {
						chatter_list_status( tmp, pixmap_status_away, "AW" );
					} else if ( ! strncasecmp( tmp2, "is back", 7 )) {
						chatter_list_status( tmp, pixmap_status_here, "" );
					}
				}
				append_to_textbox( chat_window, NULL, buf );

				break;
		case YMSG9_JOIN:
				/* fixme, need to verify what exactly constitutes a join pkt */
				if ( strcmp( ymsg9_field( "61" ), "" )) {
					/* I joined the room */
					gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
					sprintf( buf, "Current Room: %s", ymsg9_field( "104" ));
					gtk_statusbar_push( GTK_STATUSBAR(chat_status),
						st_cid, buf );
					strcpy( sess->room, ymsg9_field( "104" ));

					sprintf( buf, "You are now in %s%s%s\n",
						YAHOO_COLOR_BLUE, ymsg9_field( "104" ), ATTR_RESET );
					if ( strcmp( ymsg9_field( "105" ), "" )) {
						strcat( buf, YAHOO_COLOR_CYAN );
						strcat( buf, ymsg9_field( "105" ));
						strcat( buf, ATTR_RESET );
						strcat( buf, "\n" );
					}

					strcpy( tmp, ymsg9_field( "109" ));
					src = tmp;
					ptr = tmp2;
					while( *src ) {
						*ptr = *(src++);
						if ( *ptr == ',' ) {
							ptr++;
							*ptr = ' ';
						}
						ptr++;
					}
					*ptr = '\0';

					strcat( buf, "You see here: " );
					strcat( buf, tmp2 );
					strcat( buf, "\n" );
					append_to_textbox( chat_window, NULL, buf );

					chatter_list_populate( ymsg9_field( "109" ));

					/* little debugging code for showing capabilities */
					/*
					strcpy( tmp, ymsg9_field( "113" ));
					strcpy( tmp2, ymsg9_field( "109" ));
					ptr = tmp;
					src = tmp2;
					while( src ) {
						end = strchr( src, ',' );
						if ( end ) {
							*end = '\0';
							end++;
						}
						end = strchr( ptr, ',' );
						if ( end ) {
							*end = '\0';
							end++;
						}

						fprintf( stderr, "%-20.20s = %x\n", src, atoi( ptr ));
						fflush( stderr );
						if ( ! end ) {
							src = NULL;
						} else {
							src = src + strlen( src ) + 1;
							ptr = ptr + strlen( ptr ) + 1;
						}
					}
					*/
				} else if (( ! strcmp( ymsg9_field( "108" ), "1" )) &&
					( strcmp( ymsg9_field( "109" ), sess->user ))) {
					/* someone else joined the room while I'm in there */
					strcpy( tmp, ymsg9_field( "109" ));
					chatter_list_add( tmp );

					if (( ! ignore_check( tmp )) &&
						( ! mute_check( tmp )) &&
						( show_enters )) {
						sprintf( buf, "%s%s %senters%s the room\n",
							YAHOO_COLOR_BLUE, tmp,
							YAHOO_COLOR_GREEN, YAHOO_COLOR_BLACK );
						append_to_textbox( chat_window, NULL, buf );
					}
				}
				break;
		case YMSG9_EXIT:
				if ( strcmp( sess->user, ymsg9_field( "109"))) {
					/* someone else left */
					strcpy( tmp, ymsg9_field( "109" ));
					chatter_list_remove( tmp );

					if (( ! ignore_check( tmp )) &&
						( ! mute_check( tmp )) &&
						( show_enters )) {
						sprintf( buf, "%s%s %sleaves%s the room\n",
							YAHOO_COLOR_BLUE, tmp,
							YAHOO_COLOR_GREEN, YAHOO_COLOR_BLACK );
						append_to_textbox( chat_window, NULL, buf );
					}
				} else {
					/* packet that shows up when I leave */
				}
				break;

		case YMSG9_PM:
		case YMSG9_PM_RECV:
				strcpy( tmp, ymsg9_field( "4" ));
				strcpy( tmp2, ymsg9_field( "14" ));

				/* don't think we should get any of these, but just in cas */
				if ( ! strcmp( ymsg9_field( "49" ), "TYPING" )) {
					return;
				}

				if ( ignore_check( tmp )) {
					return;
				}

				if ( mute_check( tmp )) {
					return;
				}

				/* check the fields determining who to accept PMs from */
				accept_pm = 0;
				if (( pm_from_all ) ||
					(( pm_from_users ) &&
					 ( find_user_row( tmp ) >= 0 )) ||
					(( pm_from_friends ) &&
					 ( find_friend( tmp )))) {
					accept_pm = 1;
				}

				if ( ! accept_pm ) {
					if ( find_friend( tmp )) {
						sprintf( buf, "Ignored PM from Friend (%s%s%s)\n",
							YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK );
						append_to_textbox( chat_window, NULL, buf );
					}
					/* possibly enable this code sometime.
					else if ( find_user_row( tmp ) >= 0 ) {
						sprintf( buf, "Ignored PM from User in Room (%s%s%s)\n",
							YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK );
						append_to_textbox( chat_window, NULL, buf );
					}
					*/
					return;
				}

				if (( ptr = auto_ignore_check( tmp2 )) != NULL ) {
					if ( ! ignore_check( tmp )) {
						ignore_toggle( tmp );
					} else {
						return;
					}

					sprintf( buf, "Gyach - auto-ignored %s%s%s, regex '%s'\n",
						YAHOO_COLOR_RED, tmp, YAHOO_COLOR_BLACK, ptr );
					append_to_textbox( chat_window, NULL, buf );

					if ( ! disp_auto_ignored ) {
						return;
					}
				} 

				if (( ! pm_in_sep_windows ) ||
					( ignore_check( tmp ))) {
					if ( chat_timestamp ) {
						time_now = time( NULL );
						tm_now = localtime( &time_now );
						sprintf( buf, "%s[%d/%d/%d %02d:%02d] ",
							YAHOO_COLOR_BLACK,
							tm_now->tm_mon + 1, tm_now->tm_mday,
							tm_now->tm_year + 1900,
							tm_now->tm_hour, tm_now->tm_min );
						append_to_textbox_color( chat_window, NULL, buf );
					}

					sprintf( buf, "%s%s%s%s%s <private to %s%s%s%s%s> %s%s%s\n",
						YAHOO_COLOR_RED, YAHOO_STYLE_BOLDON, tmp,
						YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF,
						YAHOO_COLOR_BLUE, YAHOO_STYLE_BOLDON, sess->user,
						YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF,
						YAHOO_STYLE_BOLDON, tmp2, YAHOO_STYLE_BOLDOFF );

					append_to_textbox_color( chat_window, NULL, buf );
				} else {
					/* here we would send to a private pm session */
					/* duplicate to both for now for testing */
					if (( pm_lptr = find_pm_session( tmp )) != NULL ) {
						pm_sess = pm_lptr->data;
					} else {
						/* open a new window if one doesn't exist already */
						pm_sess = (PM_SESSION *)malloc( sizeof( PM_SESSION ));
						pm_sess->pm_window = create_pm_session();
						pm_sess->pm_text = lookup_widget( pm_sess->pm_window,
							"pms_text" );
						pm_sess->pm_user = strdup( tmp );
						pm_sess->typing = 0;

						sprintf( buf, "* %s - Gyach", tmp );
						gtk_window_set_title( GTK_WINDOW(pm_sess->pm_window), buf );

						pm_list = g_list_append( pm_list, pm_sess );

						/* setup the callback for url highlighting */
						gtk_signal_connect_after (GTK_OBJECT (pm_sess->pm_text),
							"motion_notify_event",
							GTK_SIGNAL_FUNC(chat_motion_notify), NULL);

						GTK_WIDGET_UNSET_FLAGS( GTK_WIDGET(pm_sess->pm_text),
							GTK_CAN_FOCUS);

						gtk_widget_set_events( GTK_WIDGET(pm_sess->pm_text),
							GDK_POINTER_MOTION_MASK
							| GDK_POINTER_MOTION_HINT_MASK );
					}
					gtk_widget_show( pm_sess->pm_window );

					if ( auto_raise_pm ) {
						gdk_window_raise( pm_sess->pm_window->window );
					}

					if ( chat_timestamp_pm ) {
						time_now = time( NULL );
						tm_now = localtime( &time_now );
						sprintf( buf, "%s[%d/%d/%d %02d:%02d] ",
							YAHOO_COLOR_BLACK,
							tm_now->tm_mon + 1, tm_now->tm_mday,
							tm_now->tm_year + 1900,
							tm_now->tm_hour, tm_now->tm_min );
						append_to_textbox_color( pm_sess->pm_window,
							pm_sess->pm_text, buf );
					}

					sprintf( buf, "%s%s%s%s%s: %s\n",
						YAHOO_COLOR_RED, YAHOO_STYLE_BOLDON, tmp,
						YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF, tmp2 );
					append_to_textbox_color( pm_sess->pm_window,
						pm_sess->pm_text, buf );
					sprintf( buf, "* %s - Gyach", tmp );
					gtk_window_set_title( GTK_WINDOW(pm_sess->pm_window), buf );
				}

				/* auto-reply if necessary */
				if (( my_status ) &&
					( auto_reply_when_away ) &&
					( ! ignore_check( tmp )) &&
					( auto_reply_msg )) {
					ymsg9_pm( tmp, auto_reply_msg );
				}
				break;

		case YMSG9_NOTIFY:
				strcpy( tmp, ymsg9_field( "4" ));

				/* message that tell us the user is typing to us or stopped */
				/* typing is stored in '49' field */
				if ( ! strcmp( ymsg9_field( "49" ), "TYPING" )) {
					/* this code broken right now, we don't seem to receive */
					/* the typing start messages */
					if (( pm_lptr = find_pm_session( tmp )) != NULL ) {
						pm_sess = pm_lptr->data;
						if ( GTK_WINDOW(pm_sess->pm_window)->title[0] == '*' ) {
							strcpy( buf, "* " );
						} else {
							buf[0] = '\0';
						}
						strcat( buf, tmp );
						if ( ! strcmp( ymsg9_field( "13" ), "1" )) {
							strcat( buf, " [TYPING]" );
						}
						strcat( buf, " - Gyach" );
						gtk_window_set_title( GTK_WINDOW(pm_sess->pm_window),
							buf );
					}
					return;
				}
				break;

		case YMSG9_BUDDY_ON:
				if ( ! strcmp( ymsg9_field( "7" ), "" )) {
					return;
				}

				if ( strcmp( ymsg9_field( "8" ), "" )) {
					/* initial buddy list of people online */
					strcpy( buf, ymsg9_field( "7"));
					ptr = buf;
					last = 0;
					while( ptr ) {
						end = strchr( ptr, ',' );
						if ( end ) {
							*end = '\0';
						} else {
							last = 1;
						}

						add_online_friend( ptr );

						if ( last ) {
							ptr = NULL;
						} else {
							ptr = end + 1;
						}
					}

					show_friends();
					return;
				}
			
				strcpy( tmp, ymsg9_field( "7" ));

				if ( ! strcmp( tmp, sess->user )) {
					return;
				}

				sprintf( buf, "%sBuddy:%s %s%s%s has logged ON.\n",
					YAHOO_COLOR_BLUE, YAHOO_COLOR_BLACK, YAHOO_COLOR_RED,
					tmp, YAHOO_COLOR_BLACK );

				display = 1;
				if ( find_online_friend( tmp )) {
					display = 0;
				} else {
					add_online_friend( tmp );
				}

				if ( display ) {
					append_to_textbox_color( chat_window, NULL, buf );
					if (( pm_lptr = find_pm_session( tmp )) != NULL ) {
						pm_sess = pm_lptr->data;
						append_to_textbox_color( pm_sess->pm_window,
							pm_sess->pm_text, buf );
					}
				}
				break;

		case YMSG9_BUDDY_OFF:
				if ( ! sess->yin.size ) {
					sess->yin.type = YMSG9_LOGOUT;
					showYahooPacket( sess );
				}
				if ( ! strcmp( ymsg9_field( "7" ), "" )) {
					return;
				}
				strcpy( tmp, ymsg9_field( "7"));

				if ( ! strcmp( tmp, sess->user )) {
					return;
				}

				sprintf( buf, "%sBuddy:%s %s%s%s has logged OFF.\n",
					YAHOO_COLOR_BLUE, YAHOO_COLOR_BLACK, YAHOO_COLOR_RED,
					tmp, YAHOO_COLOR_BLACK );

				display = 1;
				if ( find_online_friend( tmp )) {
					remove_online_friend( tmp );
				} else {
					display = 0;
				}

				if ( display ) {
					append_to_textbox_color( chat_window, NULL, buf );
					if (( pm_lptr = find_pm_session( tmp )) != NULL ) {
						pm_sess = pm_lptr->data;
						append_to_textbox_color( pm_sess->pm_window,
							pm_sess->pm_text, buf );
					}
				}
				break;

		case YMSG9_INVITE: 
				strcpy( tmp, ymsg9_field( "119" ));
				strcpy( tmp2, ymsg9_field( "104" ));

				if ( ! strcmp( tmp, "" )) {
					sprintf( buf, "%sUser has been invited to %s%s\n",
						YAHOO_COLOR_PURPLE, sess->room, YAHOO_COLOR_BLACK );
					append_to_textbox_color( chat_window, NULL, buf );
					return;
				}
			
				sprintf( buf,
					"%sYou have been invited to (%s%s%s%s%s) by %s%s%s%s%s\n",
					YAHOO_COLOR_PURPLE, YAHOO_COLOR_RED, YAHOO_STYLE_BOLDON,
					tmp2, YAHOO_STYLE_BOLDOFF, YAHOO_COLOR_PURPLE,
					YAHOO_COLOR_BLUE, YAHOO_STYLE_BOLDON, tmp,
					YAHOO_STYLE_BOLDOFF, YAHOO_COLOR_BLACK );
				if (( ignore_check( tmp )) ||
					( disp_auto_ignored ))
					append_to_textbox_color( chat_window, NULL, buf );
				break;

		case YMSG9_FILETRANSFER:
				strcpy( tmp, ymsg9_field( "4" ));

				if ( ! strcmp( tmp, "FILE_TRANSFER_SYSTEM" )) {
					/* we sent a file */
					strcpy( tmp2, ymsg9_field( "14" ));
					sprintf( buf, "%sFILE:%s %s\n",
						YAHOO_COLOR_GREEN, YAHOO_COLOR_BLACK, tmp2 );
					append_to_textbox_color( chat_window, NULL, buf );
				} else {
					/* we received a file */
					strcpy( tmp2, ymsg9_field( "20" ));
					strcpy( tmp3, ymsg9_field( "14" ));
					exp_time = atoi( ymsg9_field( "38" ));
					sprintf( buf, "%sFILE:%s %s%s has sent you a file with "
						"the message: '%s'.\n"
						"        %s\n"
						"        It is available until: %s",
						YAHOO_COLOR_GREEN, YAHOO_COLOR_BLUE, tmp, 
						YAHOO_COLOR_BLACK, tmp3, tmp2, ctime( &exp_time ));
					append_to_textbox_color( chat_window, NULL, buf );
				}
				break;

		case YMSG9_LOGOUT:
				closesocket( sess->sock );
				sess->sock = INVALID_SOCKET;
				logged_in = 0;
				set_menu_connected( 0 );

				if (io_callback_tag != 0) {
					gdk_input_remove( io_callback_tag );
					io_callback_tag = 0;
				}

				time_now = time(NULL);

				hours = (int)(( time_now - connect_time ) / 3600);
				minutes = (int)(( time_now - connect_time ) % 3600 ) / 60;
				seconds = (int)(( time_now - connect_time ) % 60 );
				connect_time = 0;

				/* show user that we aren't logged in anymore */
				sprintf( buf, "%s%s%s%s%s  Online for %d:%02d:%02d.\n\n",
					YAHOO_COLOR_BLUE, YAHOO_STYLE_BOLDON,
					"Disconnected from Chat!",
					YAHOO_STYLE_BOLDOFF, YAHOO_COLOR_BLACK,
					hours, minutes, seconds );
				append_to_textbox_color( chat_window, NULL, buf );

				gtk_statusbar_pop( GTK_STATUSBAR(chat_status), st_cid );
				gtk_statusbar_push( GTK_STATUSBAR(chat_status), st_cid,
					"Not currently connected to chat..." );

				if ( auto_reconnect ) {
					login_to_yahoo_chat();
				} 
				break;
	}

}

int build_toolbar( GtkWidget *chat_window ) {
	GtkWidget *tgl_exit;
	GtkWidget *tgl_debug;
	GtkWidget *tgl_bold;
	GtkWidget *tgl_italics;
	GtkWidget *tgl_underline;
	GtkWidget *chat_toolbar;
	GdkPixmap *st_pixmap = NULL;
	GdkBitmap *st_mask = NULL;
	GtkWidget *tmp_pixmap;
	GtkStyle *user_style = NULL;
	int tsize = 20;        /* size of toolbar buttons in pixels wide & high */

	chat_toolbar = lookup_widget( chat_window, "chat_toolbar" );

	user_style = gtk_widget_get_style( chat_toolbar );

	/* exit button */
	st_pixmap = gdk_pixmap_create_from_xpm_d(
		chat_toolbar->window, &st_mask,
		&user_style->bg[GTK_STATE_NORMAL],
		(gchar **)pixmap_exit );
	tmp_pixmap = gtk_pixmap_new( st_pixmap, st_mask );
	tgl_exit = gtk_toolbar_append_element (GTK_TOOLBAR (chat_toolbar),
		GTK_TOOLBAR_CHILD_BUTTON, NULL, " Exit ", "Exit Gyach",
		NULL, tmp_pixmap, NULL, NULL);
	gtk_widget_ref (tgl_exit);
	gtk_object_set_data_full (GTK_OBJECT (chat_window), "tgl_exit", tgl_exit,
		(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (tgl_exit);
	gtk_widget_set_usize (tgl_exit, tsize, tsize);

	/* give us a blank */
	gtk_toolbar_append_space( GTK_TOOLBAR (chat_toolbar));

	/* packet debugging button */
	st_pixmap = gdk_pixmap_create_from_xpm_d(
		chat_toolbar->window, &st_mask,
		&user_style->bg[GTK_STATE_NORMAL],
		(gchar **)pixmap_debug );
	tmp_pixmap = gtk_pixmap_new( st_pixmap, st_mask );
	tgl_debug = gtk_toolbar_append_element (GTK_TOOLBAR (chat_toolbar),
		GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL, " Debug ", "Packet Debugging",
		NULL, tmp_pixmap, NULL, NULL);
	gtk_widget_ref (tgl_debug);
	gtk_object_set_data_full (GTK_OBJECT (chat_window), "tgl_debug", tgl_debug,
		(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (tgl_debug);
	gtk_widget_set_usize (tgl_debug, tsize, tsize);

	/* give us a blank */
	gtk_toolbar_append_space( GTK_TOOLBAR (chat_toolbar));

	/* now the bold button */
	st_pixmap = gdk_pixmap_create_from_xpm_d(
		chat_toolbar->window, &st_mask,
		&user_style->bg[GTK_STATE_NORMAL],
		(gchar **)pixmap_bold );
	tmp_pixmap = gtk_pixmap_new( st_pixmap, st_mask );
	tgl_bold = gtk_toolbar_append_element (GTK_TOOLBAR (chat_toolbar),
		GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL, " B ", "Send your text in Bold",
		NULL, tmp_pixmap, NULL, NULL);
	gtk_widget_ref (tgl_bold);
	gtk_object_set_data_full (GTK_OBJECT (chat_window), "tgl_bold", tgl_bold,
		(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (tgl_bold);
	gtk_widget_set_usize (tgl_bold, tsize, tsize);

	/* now the italic button */
	st_pixmap = gdk_pixmap_create_from_xpm_d(
		chat_toolbar->window, &st_mask,
		&user_style->bg[GTK_STATE_NORMAL],
		(gchar **)pixmap_italic );
	tmp_pixmap = gtk_pixmap_new( st_pixmap, st_mask );
	tgl_italics = gtk_toolbar_append_element (GTK_TOOLBAR (chat_toolbar),
		GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL, " I ", "Send your text in italic",
		NULL, tmp_pixmap, NULL, NULL);
	gtk_widget_ref (tgl_italics);
	gtk_object_set_data_full (GTK_OBJECT (chat_window), "tgl_italics",
		tgl_italics, (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (tgl_italics);
	gtk_widget_set_usize (tgl_italics, tsize, tsize);

	st_pixmap = gdk_pixmap_create_from_xpm_d(
		chat_toolbar->window, &st_mask,
		&user_style->bg[GTK_STATE_NORMAL],
		(gchar **)pixmap_underline );
	tmp_pixmap = gtk_pixmap_new( st_pixmap, st_mask );
	tgl_underline = gtk_toolbar_append_element (GTK_TOOLBAR (chat_toolbar),
		GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL, " U ",
		"Send your text in underline", NULL, tmp_pixmap, NULL, NULL);
	gtk_widget_ref (tgl_underline);
	gtk_object_set_data_full (GTK_OBJECT (chat_window), "tgl_underline",
		tgl_underline, (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (tgl_underline);
	gtk_widget_set_usize (tgl_underline, tsize, tsize);

	/* setup some signals */
	gtk_signal_connect (GTK_OBJECT (tgl_exit), "clicked",
		GTK_SIGNAL_FUNC (on_exit_clicked), NULL);
	gtk_signal_connect (GTK_OBJECT (tgl_debug), "toggled",
		GTK_SIGNAL_FUNC (on_packet_debugging_toggled), NULL);
	gtk_signal_connect (GTK_OBJECT (tgl_bold), "toggled",
		GTK_SIGNAL_FUNC (on_tgl_bold_toggled), NULL);
	gtk_signal_connect (GTK_OBJECT (tgl_italics), "toggled",
		GTK_SIGNAL_FUNC (on_tgl_italics_toggled), NULL);
	gtk_signal_connect (GTK_OBJECT (tgl_underline), "toggled",
		GTK_SIGNAL_FUNC (on_tgl_underline_toggled), NULL);

	return( 0 );
}

GtkWidget *build_chat_window( void ) {
	GtkWidget *tmp_widget;
	GtkStyle *windowstyle;
	GdkPixmap *pm_icon;
	GdkBitmap *mask_icon;

	chat_window = create_window1();
	gtk_window_set_title( GTK_WINDOW(chat_window), "Gyach v" VERSION );

	/* set default window sizes from saved values */
	if ( remember_position )
		gtk_widget_set_uposition( chat_window, chat_window_x, chat_window_y );
	gtk_widget_set_usize( chat_window, chat_window_width, chat_window_height );
	gtk_window_set_default_size( GTK_WINDOW(chat_window), chat_window_width,
		chat_window_height );

	ct_build();
	ct_set_max_lines( buffer_size );

	tmp_widget = lookup_widget( chat_window, "hbox2" );
	gtk_widget_set_usize( tmp_widget, chat_text_width, -2 );
	tmp_widget = lookup_widget( chat_window, "scrolledwindow3" );
	gtk_widget_set_usize( tmp_widget, -2, chat_user_height );

	tmp_widget = lookup_widget( chat_window, "chat_entry" );

	gtk_widget_show( chat_window );

	gtk_clist_set_compare_func( GTK_CLIST(chat_users), clist_strcasecompare );

	/* attach our icon */
	windowstyle = gtk_widget_get_style( chat_window );
	pm_icon = gdk_pixmap_create_from_xpm_d( chat_window->window,
		&mask_icon, &(windowstyle->bg[GTK_STATE_NORMAL]),
		(gchar **)pixmap_gyach_icon );
	gdk_window_set_icon( chat_window->window, NULL, pm_icon, mask_icon );
	gdk_pixmap_unref( pm_icon );
	gdk_bitmap_unref( mask_icon );

	build_toolbar( chat_window );

	st_cid = gtk_statusbar_get_context_id( GTK_STATUSBAR(chat_status),
		"Gyach" );

	set_menu_connected( 0 );

	/* set any options that need the window created */
	set_config();

	chat_user_menu = create_user_menu();
	emote_menu = create_emote_menu();
	setup_menu = create_setup_menu();

	return( chat_window );
}


GtkWidget *build_login_window( void ) {
	GtkWidget *tmp_widget;
	GList *serverlist = NULL;

	login_window = create_login_window();

	if ( ! password )
		password = strdup( "" );

	tmp_widget = lookup_widget( login_window, "chat_username" );

	if ( login_list )
		gtk_combo_set_popdown_strings( GTK_COMBO(tmp_widget), login_list );

	tmp_widget = lookup_widget( login_window, "chat_password" );
	gtk_entry_set_text( GTK_ENTRY(tmp_widget), password );

	if ( ! favroom_list ) {
		favroom_list = g_list_append( favroom_list,
			strdup( "Linux, FreeBSD, Solaris:1" ));
		favroom_list = g_list_append( favroom_list, strdup( "Programming:1" ));
	}

	tmp_widget = lookup_widget( login_window, "chat_room" );

	gtk_combo_set_popdown_strings( GTK_COMBO(tmp_widget), favroom_list );

	/* make cs6 the default for now by putting it first */
	/* fixme, save this setting in the gyachrc file */
	/* fixme, make it auto-detect these and try going through the list */
	serverlist = g_list_append( serverlist, "scs.yahoo.com" );
	serverlist = g_list_append( serverlist, "scs-fooa.yahoo.com" );
	serverlist = g_list_append( serverlist, "scs-foob.yahoo.com" );
	serverlist = g_list_append( serverlist, "scs-fooc.yahoo.com" );
	serverlist = g_list_append( serverlist, "scs-food.yahoo.com" );
	serverlist = g_list_append( serverlist, "scs-fooe.yahoo.com" );
/*
	serverlist = g_list_append( serverlist, "cs.yahoo.com" );
	serverlist = g_list_append( serverlist, "cs8.chat.yahoo.com" );
	serverlist = g_list_append( serverlist, "cs7.chat.yahoo.com" );
	serverlist = g_list_append( serverlist, "cs6.chat.yahoo.com" );
	serverlist = g_list_append( serverlist, "cs.chat.yahoo.com" );
	serverlist = g_list_append( serverlist, "cs2.chat.yahoo.com" );
	serverlist = g_list_append( serverlist, "cs3.chat.yahoo.com" );
	serverlist = g_list_append( serverlist, "cs4.chat.yahoo.com" );
	serverlist = g_list_append( serverlist, "cs5.chat.yahoo.com" );
*/

	tmp_widget = lookup_widget( login_window, "chat_server" );
	gtk_combo_set_popdown_strings( GTK_COMBO(tmp_widget), serverlist );

	if ( login_list ) {
		strcpy( sess->user, (char*)login_list->data );
	} else {
		strcpy( sess->user, "" );
	}
	strcpy( sess->pass, password );
	strcpy( sess->room, (char*)favroom_list->data );
	strcpy( sess->host, (char*)serverlist->data );

	return( login_window );
}

int main( int argc, char **argv ) {
	char gtkrc_filename[1024];
	struct stat gtkrc_stat;
#ifdef OS_WINDOWS
	struct timeval tv;
	fd_set set;
	int ret;
#endif

	DBG( 1, "Gyach v%s\n", VERSION );
	DBG( 1, "%s\n", GYACH_URL );
	DBG( 1, "Debug Level set to: %d\n", DBG_LEVEL );

	memset(sess ,0 ,sizeof(SESSION)); 
	sess->rval = 1; 
	sess->sock = INVALID_SOCKET; 

	gtk_set_locale ();

	gtk_init (0,0);

#ifdef OS_WINDOWS
	sprintf( gtkrc_filename, "./gtkrc" );
#else
	sprintf( gtkrc_filename, "%s/.gyach/gtkrc", getenv( "HOME" ));
#endif
	if ( ! stat( gtkrc_filename, &gtkrc_stat )) {
		gtk_rc_parse( gtkrc_filename );
	}

	gyach_init();

	chat_window = build_chat_window();
	gtk_widget_show( chat_window );

	login_window = build_login_window();

	if ( ! auto_login ) {
		gtk_widget_show( login_window );
	} else {
		login_to_yahoo_chat();
	}

	gdk_threads_enter();

#ifdef OS_WINDOWS
	/* since we can't use an io callback with win32, we have to do a */
	/* main loop ourselves and check the socket with a select().     */
	while( ! sess->quit ) {
		while( gtk_events_pending()) {
			gtk_main_iteration();
		}

		/* now check for input on our socket */
		FD_ZERO( &set );
		if ( sess->sock != INVALID_SOCKET )
			FD_SET( sess->sock, &set );

		tv.tv_sec = 0;
		tv.tv_usec = 50000;
		ret = select( sess->sock + 1, &set, NULL, NULL, &tv );

		if ( ret ) {
			if ( ymsg9_recv_data()) {
				showYahooPacket( sess );
			}
		}
	}
#else
	gtk_main();
#endif

	return( 0 );
}

