/*****************************************************************************
 * yahoochat.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * Copyright (C) 2000-2002 Chris Pinkham
 * Released under the terms of the GPL.
 * *NO WARRANTY*
 *
 * cpinkham@infi.net, cpinkham@bc2va.org
 * http://www4.infi.net/~cpinkham/gyach/
 *****************************************************************************/

#include <time.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#include "yahoochat.h"

#include "gyach.h"
#include "commands.h"
#include "main.h"
#include "yahoo_md5.h"

extern SESSION *sess;

char *yahoo_status[] = {
	"Here",
	"Be Right Back",
	"Busy",
	"Not at home",
	"Not at my desk",
	"Not in the Office",
	"On the phone",
	"On vacation",
	"Out to lunch",
	"Stepped out",
	"" };


/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

int ymsg9_sess_id = 0;

int sendN(SESSION *sess ,const u_char *buf ,int cnt ,const char *info);
int recvN(SESSION *sess ,char *buf ,int cnt ,const char *info);
void append_to_textbox(GtkWidget *window, GtkWidget *textbox, char *text);


u_char *ymsg9_header( u_char *buf, u_long pkt_type ) {
	u_char *ptr = buf;

	memcpy( buf, "YMSG", 4 );					ptr += 4;  /* msg procotol */

	switch( pkt_type ) {
		case YMSG9_AWAY:
		case YMSG9_BACK:
				*((u_long *)ptr) = htonl( 0x00070000 );	ptr += 4;  
				break;
		case YMSG9_ADD_BUDDY:
		case YMSG9_REM_BUDDY:
				*((u_long *)ptr) = htonl( 0x06000000 );	ptr += 4;  
				break;
		default:
				*((u_long *)ptr) = htonl( 0x09000000 );	ptr += 4;  
				break;
	}

	/* next 4 bytes are packet length which we'll */
	/* fill in later, so set to 0 for now         */
	*((u_short *)ptr) = htons( 0x0000 );		ptr += 2;  /* data length  */

	*((u_short *)ptr) = htons( pkt_type );		ptr += 2;  /* packet type  */
	*((u_long *)ptr) = htonl( 0x00000000 );		ptr += 4;  /* status       */
	switch( pkt_type ) {
		case YMSG9_ADD_BUDDY:
		case YMSG9_REM_BUDDY:
				*((u_long *)ptr) =
					htonl( 0x00000000 );	ptr += 4;  /* Session ID   */
				break;
		default:
				*((u_long *)ptr) =
					htonl( ymsg9_sess_id );	ptr += 4;  /* Session ID   */
				break;
	}

	return( ptr );
}

void ymsg9_dump_packet( u_char *pkt ) {
	int len, type;
	int i = 0;
	int x = 0;
	int y = 0;
	char *ptr;

	DBG( 21, "ymsg9_dump_packet()\n" );

	if ( strncmp( pkt, "YMSG", 4 )) {
		printf( "ymsg9_dump_packet(), invalid packet\n" );
		return;
	}

	len = ntohs((u_short)*((u_short*)(pkt+8)));
	type = ntohs((u_short)*((u_short*)(pkt+10)));

	fprintf( stderr, "------------------------------------------------------------------------------\n" );
	fprintf( stderr, "sess id   : %u\n", ymsg9_sess_id );
	fprintf( stderr, "data len  : %d\n", len );
	fprintf( stderr, "pkt type  : 0x%x\n", type );

	for( i = 0; i < 20; i++ ) {
		fprintf( stderr, "%02x ", *(pkt+i) & 0xFF );
	}
	fprintf( stderr, "\n" );
	for( i = 0; i < 20; i++ ) {
		if ( isgraph( *(pkt+i) ) || *(pkt+i)  == ' ' ) {
			fprintf( stderr, "%c  ", *(pkt+i) );
		} else {
			fprintf( stderr, ".  " );
		}
	}
	fprintf( stderr, "\n" );

	i = 0;
	ptr = pkt + 20;
	while( i < len ) {
		if ( x < 16 ) {
			fprintf( stderr, "%02x ", *(ptr+i) & 0xFF );
			x++;
			i++;
		} else {
			fprintf( stderr, "   " );
			for( ; x > 0 ; x-- ) {
				if ( isgraph( *(ptr+i-x) ) || *(ptr+i-x)  == ' ' ) {
					fprintf( stderr, "%c", *(ptr+i-x) );
				} else {
					fprintf( stderr, "." );
				}
			}

			fprintf( stderr, "\n" );
			x = 0;
		}
	}
	for( y = x; y < 16 ; y++ ) {
		fprintf( stderr, "   " );
	}
	fprintf( stderr, "   " );
	for( ; x > 0 ; x-- ) {
		if ( isgraph( *(ptr+i-x) ) || *(ptr+i-x)  == ' ' ) {
			fprintf( stderr, "%c", *(ptr+i-x) );
		} else {
			fprintf( stderr, "." );
		}
	}

	fprintf( stderr, "\n" );
}


int ymsg9_send_packet( u_char *bufp, int size, const char *info ) {
	/* set data length in packet */
	*((u_short *)(bufp+8)) = htons( size - 20 );

	/* gaim sends this so maybe see how it works */
	/*
	*((u_long *)(bufp+12)) = htonl( my_status );
	*/

	if (ntohs((u_short)*((u_short*)(bufp+10))) != YMSG9_PING ) {
		auto_away_time = time(NULL) + ( auto_away * 60 );
	}

	/* don't try to send if we're not connected */
	if ( sess->sock == INVALID_SOCKET ) {
		return( 0 );
	}

	if ( debug_packets )
		ymsg9_dump_packet( bufp );

	/* send the packet */
	return sendN( sess, bufp, size, info );
}

int ymsg9_recv_data() {
	u_char buf[2048];

	DBG( 21, "ymsg9_recv_data()\n" );

	if ( ! recvN( sess, buf, 20, "Read Header" )) {
		return( 0 );
	}

	sess->yin.size = ntohs((u_short)*((u_short*)(buf+8)));

	if ( ! recvN( sess, buf + 20, sess->yin.size, "Read Data" )) {
		return( 0 );
	}

	sess->yin.type = ntohs((u_short)*((u_short*)(buf+10)));

	if ( sess->yin.type == YMSG9_GET_KEY ) {
		ymsg9_sess_id  = ntohl((u_long)*((u_long *)(buf+16)));
	}

	if ( debug_packets )
		ymsg9_dump_packet( buf );

	buf[sess->yin.size + 20] = '\0';

	memcpy( sess->yin.data, buf + 20, sess->yin.size );

	sess->yin.data[ sess->yin.size ] = '\0';

	return( 1 );
}

int ymsg9_request_key() {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_GET_KEY );
	int len;

	sprintf( ptr, "1%s%s%s", YMSG9_SEP, sess->user, YMSG9_SEP );
	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );
	
	return( 0 );
}

int ymsg9_login( char *key ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_LOGIN );
	int len;

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
		"0", YMSG9_SEP, sess->user, YMSG9_SEP,
		"6", YMSG9_SEP, getstr1( sess->user, sess->pass, key ), YMSG9_SEP,
		"96", YMSG9_SEP, getstr2( sess->user, sess->pass, key ), YMSG9_SEP,
		"2", YMSG9_SEP, sess->user, YMSG9_SEP,
		"1", YMSG9_SEP, sess->user, YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );
	
	return( 0 );
}

int ymsg9_online() {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_ONLINE );
	int len;

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s",
		"109", YMSG9_SEP, sess->user, YMSG9_SEP,
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"6", YMSG9_SEP, "abcde", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );
	
	return( 0 );
}

int ymsg9_join() {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_JOIN );
	int len;

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"62", YMSG9_SEP, "2", YMSG9_SEP,
		"104", YMSG9_SEP, sess->room, YMSG9_SEP,
		"129", YMSG9_SEP, "0", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );
	
	return( 0 );
}

int ymsg9_comment( char *text ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_COMMENT );
	int len;

	/* come back except from invisible status */
	if (( my_status ) &&
		( my_status != 12 )) {
		cmd_away( "0" );
	} 

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"104", YMSG9_SEP, sess->room, YMSG9_SEP,
		"117", YMSG9_SEP, text, YMSG9_SEP,
		"124", YMSG9_SEP, "1", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );

	/* fixme, need to print it myself here */
	sprintf( buf, "%s%s%s%s%s: ",
		YAHOO_STYLE_BOLDON, YAHOO_COLOR_BLUE, sess->user,
		YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF );
	append_to_textbox( chat_window, NULL, buf );
	sprintf( buf, "%s%s%s\n",
		YAHOO_COLOR_BLACK, text, YAHOO_COLOR_BLACK );
	append_to_textbox( chat_window, NULL, buf );
	
	return( 0 );
}

int ymsg9_emote( char *text ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_COMMENT );
	int len;

	/* come back except from invisible status */
	/* DON'T do this, it causes an away/back cycle since /away sends an emote
	if (( my_status ) &&
		( my_status != 12 )) {
		cmd_away( "0" );
	} 
	*/

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"104", YMSG9_SEP, sess->room, YMSG9_SEP,
		"117", YMSG9_SEP, text, YMSG9_SEP,
		"124", YMSG9_SEP, "2", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );
	
	/* fixme, need to print it myself here */
	sprintf( buf, "%s%s%s%s%s ",
		YAHOO_STYLE_BOLDON, YAHOO_COLOR_BLUE, sess->user,
		YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF );
	append_to_textbox( chat_window, NULL, buf );
	sprintf( buf, "%s%s%s\n",
		YAHOO_COLOR_PURPLE, text, YAHOO_COLOR_BLACK );
	append_to_textbox( chat_window, NULL, buf );
	
	return( 0 );
}

int ymsg9_think( char *text ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_COMMENT );
	int len;

	/* come back except from invisible status */
	if (( my_status ) &&
		( my_status != 12 )) {
		cmd_away( "0" );
	} 

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s. o O ( %s )%s%s%s%s%s",
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"104", YMSG9_SEP, sess->room, YMSG9_SEP,
		"117", YMSG9_SEP, text, YMSG9_SEP,
		"124", YMSG9_SEP, "3", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );
	
	/* fixme, need to print it myself here */
	sprintf( buf, "%s%s%s%s%s ",
		YAHOO_STYLE_BOLDON, YAHOO_COLOR_BLUE, sess->user,
		YAHOO_COLOR_BLACK, YAHOO_STYLE_BOLDOFF );
	append_to_textbox( chat_window, NULL, buf );
	sprintf( buf, "%s. o O ( %s )%s\n",
		YAHOO_COLOR_PURPLE, text, YAHOO_COLOR_BLACK );
	append_to_textbox( chat_window, NULL, buf );
	
	return( 0 );
}

int ymsg9_logout() {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_LOGOUT );
	int len;

	sprintf( ptr, "%s%s%s%s",
		"1", YMSG9_SEP, sess->user, YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );

	return( 0 );
}

int ymsg9_ping() {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_PING );
	int len;

	sprintf( ptr, "%s%s%s%s",
		"109", YMSG9_SEP, sess->user, YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );

	return( 0 );
}

char *away_msgs[] = {
	"Here",
	"Be Right Back",
	"Busy",
	"Not At Home",
	"Not At My Desk",
	"Not In The Office",
	"On The Phone",
	"On Vacation",
	"Out To Lunch",
	"Stepped Out",
	"",
	"Auto-Away",
	"Invisible",
	NULL
	};

int ymsg9_away( char *msg ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_AWAY );
	int len;
	char *mptr;
	char tmp[256];

	strcpy( tmp, msg );
	mptr = strchr( tmp, ':' );
	*mptr = '\0';
	mptr++;

	if ( ! strcmp( tmp, "10" )) {
		strcpy( tmp, "99" );
	}

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s",
		"10", YMSG9_SEP, tmp, YMSG9_SEP,
		"19", YMSG9_SEP, mptr, YMSG9_SEP,
		"47", YMSG9_SEP, "0"
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );

	/* now let the room know we're away */
	if ( atoi( tmp ) == 99 ) {
		sprintf( buf, "is away ( %s ).", mptr );
	} else {
		sprintf( buf, "is away ( %s ).", away_msgs[ atoi( msg ) ] );
	}
	ymsg9_emote( buf );

	return( 0 );
}

int ymsg9_back() {
	u_char buf[4096];
	int len;

	ymsg9_header( buf, YMSG9_BACK );
	len = 20;

	ymsg9_send_packet( buf, len, "send" );

	/* now let the room know we're back */
	sprintf( buf, "is back." );
	ymsg9_emote( buf );

	return( 0 );
}

int ymsg9_pm( char *remote_user, char *msg ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_PM );
	int len;

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"5", YMSG9_SEP, remote_user, YMSG9_SEP,
		"14", YMSG9_SEP, msg, YMSG9_SEP,
		"97", YMSG9_SEP, "1", YMSG9_SEP,
		"63", YMSG9_SEP, ";0", YMSG9_SEP,
		"64", YMSG9_SEP, "0", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );

	return( 0 );
}

int ymsg9_add_buddy( char *friend ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_ADD_BUDDY );
	int len;

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s",
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"7", YMSG9_SEP, friend, YMSG9_SEP,
		"65", YMSG9_SEP, "Friends", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );

	return( 0 );
}

int ymsg9_remove_buddy( char *friend ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_REM_BUDDY );
	int len;

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s",
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"7", YMSG9_SEP, friend, YMSG9_SEP,
		"65", YMSG9_SEP, "Friends", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );

	return( 0 );
}

int ymsg9_goto( char *friend ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_GOTO );
	int len;

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s",
		"109", YMSG9_SEP, friend, YMSG9_SEP,
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"62", YMSG9_SEP, "2", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );

	return( 0 );
}

int ymsg9_invite( char *remote_user, char *room ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_INVITE );
	int len;

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"118", YMSG9_SEP, remote_user, YMSG9_SEP,
		"104", YMSG9_SEP, room, YMSG9_SEP,
		"117", YMSG9_SEP, "", YMSG9_SEP,
		"129", YMSG9_SEP, "0", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );

	return( 0 );
}

int ymsg9_verify_server() {
	u_char buf[4096];
	int len;

	ymsg9_header( buf, YMSG9_VERIFY );
	len = 20;

	ymsg9_send_packet( buf, len, "send" );

	return( 0 );
}

int ymsg9_typing( char *remote_user, int typing ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_NOTIFY );
	int len;

	/* set the typing status field */
	*((u_long *)(buf+12)) = htonl( 0x16 );

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
		"1", YMSG9_SEP, sess->user, YMSG9_SEP,
		"4", YMSG9_SEP, sess->user, YMSG9_SEP,
		"5", YMSG9_SEP, remote_user, YMSG9_SEP,
		"13", YMSG9_SEP, typing ? "1" : "0", YMSG9_SEP,
		"14", YMSG9_SEP, " ", YMSG9_SEP,
		"49", YMSG9_SEP, "TYPING", YMSG9_SEP,
		"1002", YMSG9_SEP, "1", YMSG9_SEP
		);

	len = 20 + strlen( ptr );

	ymsg9_send_packet( buf, len, "send" );

	return( 0 );
}

int ymsg9_send_file( char *remote_user, char *msg, char *filename ) {
	u_char buf[4096];
	u_char *ptr = ymsg9_header( buf, YMSG9_FILETRANSFER );
	int len;
	int sock;
	struct sockaddr_in sa;
	struct hostent *hinfo;
	char url[1025];
	struct stat sbuf;
	int fd;
	int bytes;

	/*****************************************/
	/* FIXME: this function needs major work */
	/*****************************************/

	if ( stat( filename, &sbuf )) {
		/* file doesn't exist so exit */
fprintf( stderr, "File '%s' doesn't exist\n", filename ); fflush( stderr );
		return( 0 );
	}

	fd = open( filename, O_RDONLY, 0600 );
	if ( fd == -1 ) {
		/* can't open file */
fprintf( stderr, "Can't open File '%s'\n", filename ); fflush( stderr );
		return( 0 );
	}

	sprintf( ptr, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%d%s%s%s",
		"0", YMSG9_SEP, sess->user, YMSG9_SEP,
		"5", YMSG9_SEP, remote_user, YMSG9_SEP,
		"14", YMSG9_SEP, msg, YMSG9_SEP,
		"27", YMSG9_SEP, filename, YMSG9_SEP,
		"28", YMSG9_SEP, (int)sbuf.st_size, YMSG9_SEP,
		"29", YMSG9_SEP
		);

	len = 20 + strlen( ptr ) + sbuf.st_size;

	*((u_short *)(buf+8)) = htons( len - 20 );

	hinfo = gethostbyname( "filetransfer.msg.yahoo.com" );
	memset( &sa, 0, sizeof(sa));
	memmove((void*)&sa.sin_addr.s_addr, hinfo->h_addr, hinfo->h_length );

	sa.sin_family = AF_INET;
	sa.sin_port = htons( 80 );

	sock = socket( AF_INET, SOCK_STREAM, 6 );
	connect( sock, (struct sockaddr*)&sa, sizeof(sa));

	snprintf( url, 1024,
		"POST /notifyt HTTP/1.0\r\n"
		"Content-length: %d\n"
		"Accept: text/html\r\n"
		"Accept: text/plain\r\n"
		"User-Agent: Gyach/%s\r\n"
		"Host: %s\r\n",
		len,
		"v0.8.3", "filetransfer.msg.yahoo.com" );

	strcat( url, "Cookie: " );
	strcat( url, sess->cook );
	strcat( url, "\r\n" );
	strcat( url, "\r\n" );

	write( sock, url, strlen( url ));

	write( sock, buf, 20 + strlen( ptr ));

	/* now read the file in and send it */
	bytes = read( fd, buf, 512 );
	while( bytes ) {
		write( sock, buf, bytes );

		bytes = read( fd, buf, 512 );
	}

	close( fd );
	close( sock );

	return( 0 );
}
